import sphinxify
import typing
import yaml

from .hooks_datacfg import (
    HooksDataYaml,
    BufferType,
    ClassData,
    FunctionData,
    PropData,
    PropAccess,
    ReturnValuePolicy,
)
from .generator_data import GeneratorData, MissingReporter
from .mangle import trampoline_signature

_missing = object()

# TODO: this isn't the best solution
def _gen_int_types():
    for i in ("int", "uint"):
        for j in ("", "_fast", "_least"):
            for k in ("8", "16", "32", "64"):
                yield f"{i}{j}{k}_t"
    yield "intmax_t"
    yield "uintmax_t"


_int32_types = set(_gen_int_types())


_rvp_map = {
    ReturnValuePolicy.TAKE_OWNERSHIP: ", py::return_value_policy::take_ownership",
    ReturnValuePolicy.COPY: ", py::return_value_policy::copy",
    ReturnValuePolicy.MOVE: ", py::return_value_policy::move",
    ReturnValuePolicy.REFERENCE: ", py::return_value_policy::reference",
    ReturnValuePolicy.REFERENCE_INTERNAL: ", py::return_value_policy::reference_internal",
    ReturnValuePolicy.AUTOMATIC: "",
    ReturnValuePolicy.AUTOMATIC_REFERENCE: ", py::return_value_policy::automatic_reference",
}


class HookError(Exception):
    pass


def _using_signature(fn):
    return f"{fn['parent']['x_qualname_']}_{fn['name']}"


class Hooks:
    """
        Header2Whatever hooks used for generating C++ wrappers
    """

    _qualname_bad = ":<>="
    _qualname_trans = str.maketrans(_qualname_bad, "_" * len(_qualname_bad))

    def __init__(self, data: HooksDataYaml, casters: typing.Dict[str, str]):
        self.gendata = GeneratorData(data)
        self.rawdata = data
        self.casters = casters

        self.types = set()
        self.class_hierarchy = {}

        self.subpackages = {}

    def report_missing(self, name: str, reporter: MissingReporter):
        self.gendata.report_missing(name, reporter)

    def _add_type_caster(self, typename: str):
        # defer until the end since there's lots of duplication
        self.types.add(typename)

    def _add_subpackage(self, v, data):
        if data.subpackage:
            var = "pkg_" + data.subpackage.replace(".", "_")
            self.subpackages[data.subpackage] = var
            v["x_module_var"] = var
        else:
            v["x_module_var"] = "m"

    def _get_type_caster_includes(self):
        includes = set()
        for typename in self.types:
            tmpl_idx = typename.find("<")
            if tmpl_idx != -1:
                typename = typename[:tmpl_idx]

            header = self.casters.get(typename)
            if header:
                includes.add(header)
        return sorted(includes)

    def _set_name(self, name, data):
        if data.rename:
            return data.rename

        sp = self.rawdata.strip_prefixes
        if sp:
            for pfx in sp:
                if name.startswith(pfx):
                    name = name[len(pfx) :]
                    break

        return name

    def _resolve_default(self, fn, name):
        if isinstance(name, (int, float)):
            return str(name)
        if name in ("NULL", "nullptr"):
            return name

        # if there's a parent, look there
        parent = fn["parent"]
        if parent:
            for prop in parent["properties"]["public"]:
                if prop["name"] == name:
                    name = f"{parent['namespace']}::{parent['name']}::{name}"
        return name

    def _get_function_signature(self, fn):
        param_sig = ", ".join(
            p.get("enum", p["raw_type"]) + "&" * p["reference"] + "*" * p["pointer"]
            for p in fn["parameters"]
        )
        param_sig = param_sig.replace(" >", ">")
        if fn["const"]:
            if param_sig:
                param_sig += " [const]"
            else:
                param_sig = "[const]"

        return param_sig

    def _enum_hook(self, en, enum_data):
        ename = en.get("name")
        value_prefix = None
        if ename:
            value_prefix = enum_data.value_prefix
            if not value_prefix:
                value_prefix = ename

            en["x_name"] = self._set_name(ename, enum_data)

        for v in en["values"]:
            name = v["name"]
            if value_prefix and name.startswith(value_prefix):
                name = name[len(value_prefix) :]
                if name[0] == "_":
                    name = name[1:]
            v["x_name"] = name

    def header_hook(self, header, data):
        """Called for each header"""
        data["trampoline_signature"] = trampoline_signature
        data["using_signature"] = _using_signature

        for en in header.enums:
            en["x_namespace"] = en["namespace"]
            enum_data = self.gendata.get_enum_data(en.get("name"))
            en["data"] = enum_data

            self._add_subpackage(en, enum_data)
            self._enum_hook(en, enum_data)

        for v in header.variables:
            var_data = self.gendata.get_prop_data(v["name"])
            v["data"] = var_data
            self._add_type_caster(v["raw_type"])

        data["type_caster_includes"] = self._get_type_caster_includes()
        data["class_hierarchy"] = self.class_hierarchy
        data["subpackages"] = self.subpackages

    def _function_hook(self, fn, data: FunctionData, internal: bool = False):
        """shared with methods/functions"""

        # Python exposed function name converted to camelcase
        x_name = self._set_name(fn["name"], data)
        if not data.rename and not x_name[:2].isupper():
            x_name = x_name[0].lower() + x_name[1:]

        # if cpp_code is specified, don't release the gil unless the user
        # specifically asks for it
        if data.no_release_gil is None:
            if data.cpp_code:
                data.no_release_gil = True

        x_in_params = []
        x_out_params = []
        x_all_params = []
        x_rets = []
        x_temps = []
        x_keepalives = []

        x_genlambda = False
        x_lambda_pre = []
        x_lambda_post = []

        # Use this if one of the parameter types don't quite match
        param_override = data.param_override

        # buffers: accepts a python object that supports the buffer protocol
        #          as input. If the buffer is an 'out' buffer, then it
        #          will request a writeable buffer. Data is written by the
        #          wrapped function to that buffer directly, and the length
        #          written (if the length is a pointer) will be returned
        buffer_params = {}
        buflen_params = {}
        if data.buffers:
            for bufinfo in data.buffers:
                if bufinfo.src == bufinfo.len:
                    raise ValueError(
                        f"buffer src({bufinfo.src}) and len({bufinfo.len}) cannot be the same"
                    )
                buffer_params[bufinfo.src] = bufinfo
                buflen_params[bufinfo.len] = bufinfo

        self._add_type_caster(fn["returns"])

        is_constructor = fn.get("constructor")

        for i, p in enumerate(fn["parameters"]):

            if is_constructor and p["reference"]:
                x_keepalives.append((1, i + 2))

            if p["raw_type"] in _int32_types:
                p["fundamental"] = True
                p["unresolved"] = False

            if p["name"] == "":
                p["name"] = "param%s" % i
            p["x_type"] = p.get("enum", p["raw_type"])
            p["x_callname"] = p["name"]
            p["x_retname"] = p["name"]

            po = param_override.get(p["name"])
            if po:
                p.update(po.dict(exclude_unset=True))

            p["x_pyarg"] = 'py::arg("%(name)s")' % p

            if "default" in p:
                p["default"] = self._resolve_default(fn, p["default"])
                p["x_pyarg"] += "=" + p["default"]

            ptype = "in"

            bufinfo = buffer_params.pop(p["name"], None)
            buflen = buflen_params.pop(p["name"], None)

            if bufinfo:
                x_genlambda = True
                bname = f"__{bufinfo.src}"
                p["constant"] = 1
                p["reference"] = 1
                p["pointer"] = 0

                p["x_callname"] = f"({p['x_type']}*){bname}.ptr"
                p["x_type"] = "py::buffer"

                # this doesn't seem to be true for bytearrays, which is silly
                # x_lambda_pre.append(
                #     f'if (PyBuffer_IsContiguous((Py_buffer*){p["name"]}.ptr(), \'C\') == 0) throw py::value_error("{p["name"]}: buffer must be contiguous")'
                # )

                # TODO: check for dimensions, strides, other dangerous things

                # bufinfo was validated and converted before it got here
                if bufinfo.type is BufferType.IN:
                    ptype = "in"
                    x_lambda_pre += [f"auto {bname} = {p['name']}.request(false)"]
                else:
                    ptype = "in"
                    x_lambda_pre += [f"auto {bname} = {p['name']}.request(true)"]

                x_lambda_pre += [f"{bufinfo.len} = {bname}.size * {bname}.itemsize"]

                if bufinfo.minsz:
                    x_lambda_pre.append(
                        f'if ({bufinfo.len} < {bufinfo.minsz}) throw py::value_error("{p["name"]}: minimum buffer size is {bufinfo.minsz}")'
                    )

            elif buflen:
                if p["pointer"]:
                    p["x_callname"] = f"&{buflen.len}"
                    ptype = "out"
                else:
                    # if it's not a pointer, then the called function
                    # can't communicate through it, so ignore the parameter
                    p["x_callname"] = buflen.len
                    x_temps.append(p)
                    ptype = "ignored"

            elif p.get("force_out") or (
                p["pointer"] and not p["constant"] and p["fundamental"]
            ):
                p["x_callname"] = "&%(x_callname)s" % p
                ptype = "out"
            elif p["array"]:
                asz = p.get("array_size", 0)
                if asz:
                    p["x_type"] = "std::array<%s, %s>" % (p["x_type"], asz)
                    p["x_callname"] = "%(x_callname)s.data()" % p
                else:
                    # it's a vector
                    pass
                ptype = "out"

            if p.get("ignore"):
                pass
            else:
                x_all_params.append(p)
                if ptype == "out":
                    x_out_params.append(p)
                    x_temps.append(p)
                elif ptype == "in":
                    x_in_params.append(p)

            self._add_type_caster(p["x_type"])

            if p["constant"]:
                p["x_type"] = "const " + p["x_type"]

            p["x_type_full"] = p["x_type"]
            p["x_type_full"] += "&" * p["reference"]
            p["x_type_full"] += "*" * p["pointer"]

            p["x_decl"] = "%s %s" % (p["x_type_full"], p["name"])

        if buffer_params:
            raise ValueError(
                "incorrect buffer param names '%s'"
                % ("', '".join(buffer_params.keys()))
            )

        x_callstart = ""
        x_callend = ""
        x_wrap_return = ""
        x_return_value_policy = _rvp_map[data.return_value_policy]

        if x_out_params:
            x_genlambda = True

            # Return all out parameters
            x_rets.extend(x_out_params)

        if fn["rtnType"] != "void":
            x_callstart = "auto __ret ="
            x_rets.insert(0, dict(x_retname="__ret", x_type=fn["rtnType"]))

        if len(x_rets) == 1 and x_rets[0]["x_type"] != "void":
            x_wrap_return = "return %s;" % x_rets[0]["x_retname"]
        elif len(x_rets) > 1:
            x_wrap_return = "return std::make_tuple(%s);" % ",".join(
                [p["x_retname"] for p in x_rets]
            )

        # Temporary values to store out parameters in
        if x_temps:
            for out in reversed(x_temps):
                x_lambda_pre.insert(0, "%(x_type)s %(name)s = 0" % out)

        # Rename functions
        if data.rename:
            x_name = data.rename
        elif data.internal or internal:
            x_name = "_" + x_name
        elif fn["constructor"]:
            x_name = "__init__"

        doc = ""
        doc_quoted = ""

        if data.doc is not None:
            doc = data.doc
        elif "doxygen" in fn:
            doc = fn["doxygen"]
            doc = sphinxify.process_raw(doc)

        if doc:
            # TODO
            doc = doc.replace("\\", "\\\\").replace('"', '\\"')
            doc_quoted = doc.splitlines(keepends=True)
            doc_quoted = ['"%s"' % (dq.replace("\n", "\\n"),) for dq in doc_quoted]

        if data.keepalive is not None:
            x_keepalives = data.keepalive

        # if "hook" in data:
        #     eval(data["hook"])(fn, data)

        # bind new attributes to the function definition
        # -> previously used locals(), but this is more explicit
        #    and easier to not mess up
        fn.update(
            dict(
                data=data,
                # transforms
                x_name=x_name,
                x_all_params=x_all_params,
                x_in_params=x_in_params,
                x_out_params=x_out_params,
                x_rets=x_rets,
                x_keepalives=x_keepalives,
                x_return_value_policy=x_return_value_policy,
                # lambda generation
                x_genlambda=x_genlambda,
                x_callstart=x_callstart,
                x_lambda_pre=x_lambda_pre,
                x_lambda_post=x_lambda_post,
                x_callend=x_callend,
                x_wrap_return=x_wrap_return,
                # docstrings
                x_doc=doc,
                x_doc_quoted=doc_quoted,
            )
        )

    def function_hook(self, fn, data):
        if fn.get("operator"):
            fn["data"] = FunctionData(ignore=True)
            return

        signature = self._get_function_signature(fn)
        data = self.gendata.get_function_data(fn, signature)
        if data.ignore:
            fn["data"] = data
            return

        self._add_subpackage(fn, data)
        self._function_hook(fn, data)

    def class_hook(self, cls, data):

        if cls["parent"] is not None and cls["access_in_parent"] == "private":
            cls["data"] = ClassData(ignore=True)
            return

        cls_name = cls["name"]
        cls_key = cls_name
        c = cls
        while c["parent"]:
            c = c["parent"]
            cls_key = c["name"] + "::" + cls_key

        class_data = self.gendata.get_class_data(cls_key)
        cls["data"] = class_data

        if class_data.ignore:
            return

        self._add_subpackage(cls, class_data)

        # fix enum paths
        for e in cls["enums"]["public"]:
            e["x_namespace"] = e["namespace"] + "::" + cls_name + "::"
            enum_data = self.gendata.get_cls_enum_data(
                e.get("name"), cls_key, class_data
            )
            e["data"] = enum_data

            self._enum_hook(e, enum_data)

        # update inheritance
        for base in cls["inherits"]:
            bqual = class_data.base_qualnames.get(base["class"])
            if bqual:
                base["x_qualname"] = bqual
            elif "::" not in base["class"]:
                base["x_qualname"] = f'{cls["namespace"]}::{base["class"]}'
            else:
                base["x_qualname"] = base["class"]

            base["x_qualname_"] = base["x_qualname"].translate(self._qualname_trans)

        ignored_bases = {ib: True for ib in class_data.ignored_bases}

        cls["x_inherits"] = [
            base
            for base in cls["inherits"]
            if not ignored_bases.pop(base["class"], None)
        ]
        if ignored_bases:
            bases = ", ".join(base["class"] for base in cls["inherits"])
            invalid_bases = ", ".join(ignored_bases.keys())
            raise ValueError(
                f"{cls_name}: ignored_bases contains non-existant bases "
                + f"{invalid_bases}; valid bases are {bases}"
            )

        cls_qualname = cls["namespace"] + "::" + cls_name
        cls["x_qualname"] = cls_qualname
        cls["x_qualname_"] = cls_qualname.translate(self._qualname_trans)

        self.class_hierarchy[cls_qualname] = [
            base["x_qualname"] for base in cls["x_inherits"]
        ] + class_data.force_depends

        has_constructor = False
        is_polymorphic = class_data.is_polymorphic

        # bad assumption? yep
        if cls["inherits"]:
            is_polymorphic = True

        for access in ("public", "protected", "private"):

            for fn in cls["methods"][access]:
                if fn["constructor"]:
                    has_constructor = True
                if fn["override"] or fn["virtual"]:
                    is_polymorphic = True

                # Ignore operators, move constructors, copy constructors
                if (
                    fn.get("operator")
                    or fn.get("destructor")
                    or (
                        fn.get("constructor")
                        and fn["parameters"]
                        and fn["parameters"][0]["class"] is cls
                    )
                ):
                    fn["data"] = FunctionData(ignore=True)
                    continue

                if access != "private":
                    internal = access != "public"

                    signature = self._get_function_signature(fn)
                    method_data = self.gendata.get_function_data(
                        fn, signature, cls_key, class_data
                    )
                    if method_data.ignore:
                        fn["data"] = method_data
                        continue

                    try:
                        self._function_hook(
                            fn, method_data, internal=internal,
                        )
                    except Exception as e:
                        raise HookError(f"{cls_key}::{fn['name']}") from e

        has_trampoline = (
            is_polymorphic and not cls["final"] and not class_data.force_no_trampoline
        )
        for access in ("public", "protected", "private"):
            # class attributes
            for v in cls["properties"][access]:
                if access in "private" or (
                    access == "protected" and not has_trampoline
                ):
                    v["data"] = PropData(ignore=True)
                    continue

                prop_name = v["name"]
                propdata = self.gendata.get_cls_prop_data(
                    prop_name, cls_key, class_data
                )
                self._add_type_caster(v["raw_type"])
                v["data"] = propdata
                if propdata.rename:
                    v["x_name"] = propdata.rename
                else:
                    v["x_name"] = v["name"] if access == "public" else "_" + v["name"]

                if propdata.access == PropAccess.AUTOMATIC:
                    # Properties that aren't fundamental or a reference are readonly unless
                    # overridden by the hook configuration
                    x_readonly = not v["fundamental"] and not v["reference"]
                elif propdata.access == PropAccess.READONLY:
                    x_readonly = True
                else:
                    x_readonly = False

                v["x_readonly"] = x_readonly

        cls["x_has_trampoline"] = has_trampoline
        if cls["x_has_trampoline"]:
            cls["x_trampoline_name"] = f"rpygen::Py{cls['x_qualname_']}<{cls_name}>"
        cls["x_has_constructor"] = has_constructor
        cls["x_varname"] = "cls_" + cls_name
        cls["x_name"] = self._set_name(cls_name, class_data)



import argparse
import glob
import inspect
from os.path import basename, dirname, exists, join, relpath, splitext
import subprocess
import sys

from .setup import Setup
from .generator_data import MissingReporter


def get_setup() -> Setup:
    s = Setup()
    s.prepare()
    return s


class BuildDep:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "build-dep", help="Install build dependencies", parents=[parent_parser],
        )
        parser.add_argument("--install", help="Actually do it", action="store_true")
        return parser

    def run(self, args):
        s = get_setup()
        requirements = s.pyproject.get("build-system", {}).get("requires", [])
        requirements.extend(s.setup_kwargs.get("install_requires", ""))
        requirements.append("wheel")

        pipargs = [
            sys.executable,
            "-m",
            "pip",
            "--disable-pip-version-check",
            "install",
        ]
        pipargs.extend(requirements)
        print(" ".join(pipargs))

        if args.install:
            p = subprocess.run(pipargs)
            return p.returncode


class GenCreator:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "create-gen",
            help="Create YAML files from parsed header files",
            parents=[parent_parser],
        )
        parser.add_argument(
            "--write", help="Write to files if they don't exist", action="store_true"
        )
        parser.add_argument("--strip-prefixes", action="append")

        return parser

    def run(self, args):

        pfx = ""
        if args.strip_prefixes:
            pfx = "strip_prefixes:\n- " + "\n- ".join(args.strip_prefixes) + "\n\n"

        s = get_setup()
        for wrapper in s.wrappers:
            reporter = MissingReporter()
            wrapper.on_build_gen("", reporter)

            nada = True
            for name, report in reporter.as_yaml():
                report = f"---\n\n{pfx}{report}"

                nada = False
                if args.write:
                    if not exists(name):
                        print("Writing", name)
                        with open(name, "w") as fp:
                            fp.write(report)
                    else:
                        print(name, "already exists!")

                print("===", name, "===")
                print(report)

            if nada:
                print("Nothing to do!")


class HeaderScanner:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "scan-headers",
            help="Generate a list of headers in TOML form",
            parents=[parent_parser],
        )
        return parser

    def run(self, args):
        s = get_setup()
        for wrapper in s.wrappers:
            for incdir in wrapper.get_include_dirs():
                files = list(
                    sorted(
                        relpath(f, incdir)
                        for f in glob.glob(join(incdir, "**", "*.h"), recursive=True)
                    )
                )

                print("generate = [")
                lastdir = None
                for f in files:
                    if "rpygen" not in f:
                        thisdir = dirname(f)
                        if lastdir is None:
                            if thisdir:
                                print("    #", thisdir)
                        elif lastdir != thisdir:
                            print()
                            if thisdir:
                                print("    #", thisdir)
                        lastdir = thisdir

                        base = splitext(basename(f))[0]
                        print(f'    {{ {base} = "{f}" }},')
                print("]")


class ImportCreator:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "create-imports",
            help="Generate suitable imports for a module",
            parents=[parent_parser],
        )
        parser.add_argument("base", help="Ex: wpiutil")
        parser.add_argument("compiled", help="Ex: wpiutil._impl.wpiutil")
        return parser

    def run(self, args):
        # Runtime Dependency Check
        try:
            import black
        except:
            print("Error, The following module is required to run this tool: black")
            exit(1)

        # TODO: could probably generate this from parsed code, but seems hard
        ctx = {}
        exec(f"from {args.compiled} import *", {}, ctx)

        relimport = "." + ".".join(
            args.compiled.split(".")[len(args.base.split(".")) :]
        )

        stmt = inspect.cleandoc(
            f"""

            # autogenerated by 'robotpy-build create-imports {args.base} {args.compiled}'
            from {relimport} import {','.join(sorted(ctx.keys()))}
            __all__ = ["{'", "'.join(sorted(ctx.keys()))}"]
        
        """
        )

        print(
            subprocess.check_output(
                ["black", "-", "-q"], input=stmt.encode("utf-8")
            ).decode("utf-8")
        )


def main():

    parser = argparse.ArgumentParser(prog="robotpy-build")
    parent_parser = argparse.ArgumentParser(add_help=False)
    subparsers = parser.add_subparsers(dest="cmd")
    subparsers.required = True

    for cls in (BuildDep, GenCreator, HeaderScanner, ImportCreator):
        cls.add_subparser(parent_parser, subparsers).set_defaults(cls=cls)

    args = parser.parse_args()
    cmd = args.cls()
    retval = cmd.run(args)

    if retval is False:
        retval = 1
    elif retval is True:
        retval = 0
    elif isinstance(retval, int):
        pass
    else:
        retval = 0

    exit(retval)



import os
from os.path import join
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import setuptools
import tempfile

from ..platforms import get_platform

# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        fname = join(tmpdir, "test.cpp")
        with open(fname, "w") as fp:
            fp.write("int main (int argc, char **argv) { return 0; }")
        try:
            compiler.compile([fname], output_dir=tmpdir, extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler, pfx, sep="="):
    """Return the -std=c++[11/14/17] compiler flag.
    The newer version is prefered over c++11 (when it is available).
    """

    flags = [f"{pfx}std{sep}c++17", f"{pfx}std{sep}c++14", f"{pfx}std{sep}c++11"]

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError("Unsupported compiler -- at least C++11 support is needed!")


def get_opts(typ):
    c_opts = {"msvc": ["/EHsc"], "unix": []}
    l_opts = {"msvc": [], "unix": []}

    platform = get_platform()
    if platform.os == "osx":
        darwin_opts = ["-stdlib=libc++", "-mmacosx-version-min=10.14"]
        c_opts["unix"] += darwin_opts
        l_opts["unix"] += darwin_opts

    return c_opts.get(typ, []), l_opts.get(typ, [])


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts, link_opts = get_opts(ct)

        if ct == "unix":
            opts.append("-s")  # strip
            opts.append("-g0")  # remove debug symbols
            opts.append(cpp_flag(self.compiler, "-"))
            if has_flag(self.compiler, "-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")
        elif ct == "msvc":
            opts.append(cpp_flag(self.compiler, "/", ":"))
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts

        # self._gather_global_includes()

        build_ext.build_extensions(self)

    def run(self):

        # files need to be generated before building can occur
        self.run_command("build_gen")

        build_ext.run(self)


if os.environ.get("RPYBUILD_PARALLEL") == "1":
    # don't enable this hack by default, because not really sure of the
    # ramifications -- however, it's really useful for development
    #
    # .. the real answer to this is cmake o_O

    # monkey-patch for parallel compilation
    # -> https://stackoverflow.com/questions/11013851/speeding-up-build-process-with-distutils/13176803#13176803
    def parallelCCompile(
        self,
        sources,
        output_dir=None,
        macros=None,
        include_dirs=None,
        debug=0,
        extra_preargs=None,
        extra_postargs=None,
        depends=None,
    ):
        # those lines are copied from distutils.ccompiler.CCompiler directly
        macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
            output_dir, macros, include_dirs, sources, depends, extra_postargs
        )
        cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)
        # parallel code
        import multiprocessing
        import multiprocessing.pool

        N = multiprocessing.cpu_count()

        def _single_compile(obj):
            try:
                src, ext = build[obj]
            except KeyError:
                return
            self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)

        # convert to list, imap is evaluated on-demand
        list(multiprocessing.pool.ThreadPool(N).imap(_single_compile, objects))
        return objects

    import distutils.ccompiler

    distutils.ccompiler.CCompiler.compile = parallelCCompile



import glob
import json
import inspect
import os
from os.path import (
    abspath,
    basename,
    dirname,
    exists,
    isdir,
    join,
    normpath,
    relpath,
    sep,
    splitext,
)
import posixpath
import sys
import shutil
import toposort
from typing import Optional, List
import yaml

from header2whatever.config import Config
from header2whatever.parse import ConfigProcessor

from setuptools import Extension

from .devcfg import get_dev_config
from .pyproject_configs import WrapperConfig, MavenLibDownload
from .generator_data import MissingReporter
from .hooks import Hooks
from .hooks_datacfg import HooksDataYaml
from .download import download_and_extract_zip


class Wrapper:
    """
        Wraps downloading bindings and generating them
    """

    # Used during preprocessing
    # -> should we change this based on what flags the compiler supports?
    _cpp_version = "__cplusplus 201703L"

    def __init__(self, package_name, cfg: WrapperConfig, setup):

        self.package_name = package_name
        self.cfg = cfg

        self.setup_root = setup.root
        self.root = join(setup.root, *package_name.split("."))

        # must match PkgCfg.name
        self.name = cfg.name

        # Compute the extension name, even if we don't create one
        extname = cfg.extension
        if not extname:
            extname = f"_{cfg.name}"

        # must match PkgCfg.libinit_import
        if cfg.libinit:
            libinit_py = cfg.libinit
            if libinit_py == "__init__.py":
                self.libinit_import = package_name
            else:
                pkg = splitext(self.cfg.libinit)[0]
                self.libinit_import = f"{package_name}.{pkg}"
        else:
            libinit_py = f"_init{extname}.py"
            self.libinit_import = f"{package_name}._init{extname}"

        self.libinit_import_py = join(self.root, libinit_py)

        self.platform = setup.platform
        self.pkgcfg = setup.pkgcfg

        # Used by pkgcfg
        self.depends = self.cfg.depends

        # Files that are generated AND need to be in the final wheel. Used by build_py
        self.generated_files = []

        self._all_deps = None

        self.extension = None
        if self.cfg.sources or self.cfg.generate:
            define_macros = [("RPYBUILD_MODULE_NAME", extname)] + [
                tuple(d.split(" ")) for d in self.platform.defines
            ]
            define_macros += [tuple(m.split(" ", 1)) for m in self.cfg.pp_defines]

            # extensions just hold data about what to actually build, we can
            # actually modify extensions all the way up until the build
            # really happens
            extname_full = f"{self.package_name}.{extname}"
            self.extension = Extension(
                extname_full,
                self.cfg.sources,
                define_macros=define_macros,
                language="c++",
            )

        if self.cfg.generate and not self.cfg.generation_data:
            raise ValueError(
                "generation_data must be specified when generate is specified"
            )

        # Setup an entry point (written during build_clib)
        entry_point = f"{self.cfg.name} = {self.package_name}.pkgcfg"

        setup_kwargs = setup.setup_kwargs
        ep = setup_kwargs.setdefault("entry_points", {})
        ep.setdefault("robotpybuild", []).append(entry_point)

        self.incdir = join(self.root, "include")
        self.rpy_incdir = join(self.root, "rpy-include")

        self.dev_config = get_dev_config(self.name)

    def _dl_url(self, classifier):
        # TODO: support development against locally installed things?
        dl = self.cfg.maven_lib_download
        repo_url = dl.repo_url
        grp = dl.group_id.replace(".", "/")
        art = dl.artifact_id
        ver = dl.version

        return f"{repo_url}/{grp}/{art}/{ver}/{art}-{ver}-{classifier}.zip"

    def _extract_zip_to(self, thing, dst, cache):
        download_and_extract_zip(self._dl_url(thing), to=dst, cache=cache)

    def _add_generated_file(self, fullpath):
        if not isdir(fullpath):
            self.generated_files.append(relpath(fullpath, self.root))

    # pkgcfg interface
    def get_include_dirs(self) -> Optional[List[str]]:
        includes = [self.incdir, self.rpy_incdir]
        for h in self.cfg.extra_includes:
            includes.append(join(self.setup_root, normpath(h)))
        return includes

    def get_library_dirs(self) -> Optional[List[str]]:
        if self.get_library_names():
            return [join(self.root, "lib")]
        return []

    def get_library_names(self) -> Optional[List[str]]:
        mcfg = self.cfg.maven_lib_download
        if mcfg:
            if mcfg.use_sources:
                return []
            elif mcfg.libs is None:
                return [mcfg.artifact_id]
            else:
                return mcfg.libs
        else:
            return []

    def get_dlopen_library_names(self) -> Optional[List[str]]:
        mcfg = self.cfg.maven_lib_download
        if mcfg and mcfg.dlopenlibs:
            return mcfg.dlopenlibs
        else:
            return []

    def get_type_casters(self, casters):
        for header, types in self.cfg.type_casters.items():
            for typ in types:
                casters[typ] = header

    def all_deps(self):
        if self._all_deps is None:
            self._all_deps = self.pkgcfg.get_all_deps(self.name)
        return self._all_deps

    def _all_includes(self, include_rpyb):
        includes = self.get_include_dirs()
        for dep in self.all_deps():
            includes.extend(dep.get_include_dirs())
        if include_rpyb:
            includes.extend(self.pkgcfg.get_pkg("robotpy-build").get_include_dirs())
        return includes

    def _all_library_dirs(self):
        libs = self.get_library_dirs()
        for dep in self.cfg.depends:
            libs.extend(self.pkgcfg.get_pkg(dep).get_library_dirs())
        return libs

    def _all_library_names(self):
        libs = list(
            set(self.get_library_names()) | set(self.get_dlopen_library_names())
        )
        for dep in self.cfg.depends:
            libs.extend(self.pkgcfg.get_pkg(dep).get_library_names())
        return list(reversed(libs))

    def _all_casters(self):
        casters = {}
        for dep in self.all_deps():
            dep.get_type_casters(casters)
        self.pkgcfg.get_pkg("robotpy-build").get_type_casters(casters)
        self.get_type_casters(casters)

        # add non-namespaced versions of all casters
        # -> in theory this could lead to a conflict, but
        #    let's see how it works in practice?
        for k, v in list(casters.items()):
            k = k.split("::")[-1]
            casters[k] = v
        return casters

    def on_build_dl(self, cache: str, srcdir: str):

        pkgcfgpy = join(self.root, "pkgcfg.py")
        srcdir = join(srcdir, self.name)

        try:
            os.unlink(self.libinit_import_py)
        except OSError:
            pass

        try:
            os.unlink(pkgcfgpy)
        except OSError:
            pass

        libnames_full = []
        dlcfg = self.cfg.maven_lib_download
        if dlcfg:
            libnames_full = self._clean_and_download(dlcfg, cache, srcdir)

        self._write_libinit_py(libnames_full)
        self._write_pkgcfg_py(pkgcfgpy)

    def _clean_and_download(
        self, dlcfg: MavenLibDownload, cache: str, srcdir: str
    ) -> List[str]:

        libdir = join(self.root, "lib")
        incdir = join(self.root, "include")

        # Remove downloaded/generated artifacts first
        shutil.rmtree(libdir, ignore_errors=True)
        shutil.rmtree(incdir, ignore_errors=True)
        shutil.rmtree(srcdir, ignore_errors=True)

        # grab headers
        self._extract_zip_to("headers", incdir, cache)

        if dlcfg.use_sources:
            self._extract_zip_to(dlcfg.sources_classifier, srcdir, cache)
            if dlcfg.sources:
                sources = [join(srcdir, normpath(s)) for s in dlcfg.sources]
                self.extension.sources.extend(sources)
            return []
        elif dlcfg.sources is not None:
            raise ValueError("sources must be None if use_sources is False!")

        libnames = self.get_library_names()
        dlopen_libnames = self.get_dlopen_library_names()

        libnames = [lib for lib in libnames if lib not in dlopen_libnames]
        libnames_full = []

        if libnames or dlopen_libnames:
            libext = dlcfg.libexts.get(self.platform.libext, self.platform.libext)
            linkext = dlcfg.linkexts.get(self.platform.linkext, self.platform.linkext)

            libnames_full = [
                f"{self.platform.libprefix}{lib}{libext}" for lib in libnames
            ]
            libnames_full += [
                f"{self.platform.libprefix}{lib}{libext}" for lib in dlopen_libnames
            ]

            extract_names = libnames_full[:]

            if libext != linkext:
                extract_names += [
                    f"{self.platform.libprefix}{lib}{linkext}" for lib in libnames
                ]

            os.makedirs(libdir)
            to = {
                posixpath.join(
                    self.platform.os, self.platform.arch, "shared", libname
                ): join(libdir, libname)
                for libname in extract_names
            }

            self._extract_zip_to(f"{self.platform.os}{self.platform.arch}", to, cache)

        for f in glob.glob(join(glob.escape(incdir), "**"), recursive=True):
            self._add_generated_file(f)

        for f in glob.glob(join(glob.escape(libdir), "**"), recursive=True):
            self._add_generated_file(f)

        return libnames_full

    def _write_libinit_py(self, libnames):

        # This file exists to ensure that any shared library dependencies
        # are loaded for the compiled extension

        init = inspect.cleandoc(
            """
        
        # This file is automatically generated, DO NOT EDIT
        # fmt: off

        from os.path import abspath, join, dirname
        _root = abspath(dirname(__file__))

        ##IMPORTS##

        """
        )

        init += "\n"

        if libnames:
            init += "from ctypes import cdll\n\n"
            for libname in libnames:
                init += f'_lib = cdll.LoadLibrary(join(_root, "lib", "{libname}"))\n'

        imports = []
        for dep in self.cfg.depends:
            pkg = self.pkgcfg.get_pkg(dep)
            if pkg.libinit_import:
                imports.append(pkg.libinit_import)

        if imports:
            imports = "# runtime dependencies\nimport " + "\nimport ".join(imports)
        else:
            imports = ""

        init = init.replace("##IMPORTS##", imports)

        with open(self.libinit_import_py, "w") as fp:
            fp.write(init)

        self._add_generated_file(self.libinit_import_py)

    def _write_pkgcfg_py(self, fname):

        library_dirs = "[]"
        library_names = self.get_library_names()
        if library_names:
            library_dirs = '[join(_root, "lib")]'

        # write pkgcfg.py
        pkgcfg = inspect.cleandoc(
            f"""
        # fmt: off
        # This file is automatically generated, DO NOT EDIT

        from os.path import abspath, join, dirname
        _root = abspath(dirname(__file__))

        libinit_import = "{self.libinit_import}"
        depends = {repr(self.cfg.depends)}

        def get_include_dirs():
            return [join(_root, "include"), join(_root, "rpy-include")##EXTRAINCLUDES##]

        def get_library_dirs():
            return {library_dirs}
        
        def get_library_names():
            return {repr(library_names)}
        """
        )

        extraincludes = ""
        if self.cfg.extra_includes:
            # these are relative to the root of the project, need
            # to resolve the path relative to the package
            pth = join(*self.package_name.split("."))

            for h in self.cfg.extra_includes:
                h = '", "'.join(relpath(normpath(h), pth).split(sep))
                extraincludes += f', join(_root, "{h}")'

        pkgcfg = pkgcfg.replace("##EXTRAINCLUDES##", extraincludes)

        type_casters = {}
        self.get_type_casters(type_casters)
        if type_casters:
            pkgcfg += (
                f"\n\ndef get_type_casters(casters):\n"
                f"    casters.update({repr(type_casters)})\n"
            )

        with open(fname, "w") as fp:
            fp.write(pkgcfg)

        self._add_generated_file(fname)

    def _load_generation_data(self, datafile):
        with open(datafile) as fp:
            data = yaml.safe_load(fp)

        if data is None:
            data = {}

        return HooksDataYaml(**data)

    def on_build_gen(
        self, cxx_gen_dir, missing_reporter: Optional[MissingReporter] = None
    ):

        if not self.cfg.generate:
            return

        cxx_gen_dir = join(cxx_gen_dir, self.name)

        if missing_reporter:
            report_only = True
        else:
            report_only = False
            missing_reporter = MissingReporter()

        thisdir = abspath(dirname(__file__))

        hppoutdir = join(self.rpy_incdir, "rpygen")
        tmpl_dir = join(thisdir, "templates")
        cpp_tmpl = join(tmpl_dir, "gen_pybind11.cpp.j2")
        hpp_tmpl = join(tmpl_dir, "gen_cls_trampoline.hpp.j2")
        classdeps_tmpl = join(tmpl_dir, "gen_classdeps.json.j2")

        pp_includes = self._all_includes(False)

        # TODO: only regenerate files if the generated files
        #       have changed
        if not report_only:

            if self.dev_config.only_generate is None:
                shutil.rmtree(cxx_gen_dir, ignore_errors=True)
                shutil.rmtree(hppoutdir, ignore_errors=True)

            os.makedirs(cxx_gen_dir, exist_ok=True)
            os.makedirs(hppoutdir, exist_ok=True)

        per_header = False
        data_fname = self.cfg.generation_data
        if self.cfg.generation_data:
            datapath = join(self.setup_root, normpath(self.cfg.generation_data))
            per_header = isdir(datapath)
            if not per_header:
                data = self._load_generation_data(datapath)
        else:
            data = HooksDataYaml()

        sources = self.cfg.sources[:]
        pp_defines = [self._cpp_version] + self.platform.defines + self.cfg.pp_defines
        casters = self._all_casters()

        # These are written to file to make it easier for dev mode to work
        classdeps = {}

        processor = ConfigProcessor(tmpl_dir)

        if self.dev_config.only_generate is not None:
            only_generate = {n: True for n in self.dev_config.only_generate}
        else:
            only_generate = None

        generation_search_path = [self.root,] + self._all_includes(False)

        for gen in self.cfg.generate:
            for name, header in gen.items():

                header = normpath(header)
                for path in generation_search_path:
                    header_path = join(path, header)
                    if exists(header_path):
                        break
                else:
                    print(generation_search_path)
                    raise ValueError("could not find " + header)

                if report_only:
                    templates = []
                    class_templates = []
                else:
                    cpp_dst = join(cxx_gen_dir, f"{name}.cpp")
                    sources.append(cpp_dst)
                    classdeps_dst = join(cxx_gen_dir, f"{name}.json")
                    classdeps[name] = classdeps_dst

                    hpp_dst = join(
                        hppoutdir,
                        "{{ cls['namespace'] | replace(':', '_') }}__{{ cls['name'] }}.hpp",
                    )

                    templates = [
                        {"src": cpp_tmpl, "dst": cpp_dst},
                        {"src": classdeps_tmpl, "dst": classdeps_dst},
                    ]
                    class_templates = [{"src": hpp_tmpl, "dst": hpp_dst}]

                if only_generate is not None and not only_generate.pop(name, False):
                    continue

                if per_header:
                    data_fname = join(datapath, name + ".yml")
                    if not exists(data_fname):
                        print("WARNING: could not find", data_fname)
                        data = HooksDataYaml()
                    else:
                        data = self._load_generation_data(data_fname)

                # for each thing, create a h2w configuration dictionary
                cfgd = {
                    # generation code depends on this being just one header!
                    "headers": [header_path],
                    "templates": templates,
                    "class_templates": class_templates,
                    "preprocess": True,
                    "pp_retain_all_content": False,
                    "pp_include_paths": pp_includes,
                    "pp_defines": pp_defines,
                    "vars": {"mod_fn": name},
                }

                cfg = Config(cfgd)
                cfg.validate()
                cfg.root = self.incdir

                hooks = Hooks(data, casters)
                processor.process_config(cfg, data, hooks)

                hooks.report_missing(data_fname, missing_reporter)

        if only_generate:
            unused = ", ".join(sorted(only_generate))
            # raise ValueError(f"only_generate specified unused headers! {unused}")
            # TODO: make this a warning

        if not report_only:
            for name, contents in missing_reporter.as_yaml():
                print("WARNING: some items not in generation yaml for", basename(name))
                print(contents)

        # generate an inline file that can be included + called
        if not report_only:
            self._write_wrapper_hpp(cxx_gen_dir, classdeps)
            gen_includes = [cxx_gen_dir]
        else:
            gen_includes = []

        # Add the root to the includes (but only privately)
        root_includes = [self.root]

        # update the build extension so that build_ext works
        self.extension.sources = sources
        # use normpath to get rid of .. otherwise gcc is weird
        self.extension.include_dirs = [
            normpath(p)
            for p in (self._all_includes(True) + gen_includes + root_includes)
        ]
        self.extension.library_dirs = self._all_library_dirs()
        self.extension.libraries = self._all_library_names()

        for f in glob.glob(join(glob.escape(hppoutdir), "*.hpp")):
            self._add_generated_file(f)

    def _write_wrapper_hpp(self, outdir, classdeps):

        decls = []
        calls = []

        def _clean(n):
            tmpl_idx = n.find("<")
            if tmpl_idx != -1:
                n = n[:tmpl_idx]
            return n

        # Need to ensure that wrapper initialization is called in base order
        # so we have to toposort it here. The data is written at gen time
        # to JSON files
        types2name = {}
        types2deps = {}
        ordering = []

        for name, jsonfile in classdeps.items():
            with open(jsonfile) as fp:
                dep = json.load(fp)

            # make sure objects without classes are also included!
            if not dep:
                ordering.append(name)

            for clsname, bases in dep.items():
                clsname = _clean(clsname)
                if clsname in types2name:
                    raise ValueError(f"duplicate class {clsname}")
                types2name[clsname] = name
                types2deps[clsname] = [_clean(base) for base in bases]

        to_sort = {}
        for clsname, bases in types2deps.items():
            clsname = types2name[clsname]
            deps = to_sort.setdefault(clsname, set())
            for base in bases:
                base = types2name.get(base)
                if base and base != clsname:
                    deps.add(base)

        ordering.extend(toposort.toposort_flatten(to_sort, sort=True))

        for name in ordering:
            decls.append(f"void init_{name}(py::module &m);")
            calls.append(f"    init_{name}(m);")

        content = (
            inspect.cleandoc(
                """

        // This file is autogenerated, DO NOT EDIT
        #pragma once
        #include <robotpy_build.h>

        // forward declarations
        ##DECLS##

        static void initWrapper(py::module &m) {
        ##CALLS##
        }
        
        """
            )
            .replace("##DECLS##", "\n".join(decls))
            .replace("##CALLS##", "\n".join(calls))
        )

        with open(join(outdir, "rpygen_wrapper.hpp"), "w") as fp:
            fp.write(content)



import os
from os.path import abspath, exists, join
from setuptools import find_packages, setup as _setup
from setuptools import Extension
from setuptools_scm import get_version
import toml

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = False


except ImportError:
    bdist_wheel = None


from .command.build_py import BuildPy
from .command.build_dl import BuildDl
from .command.build_gen import BuildGen
from .command.build_ext import BuildExt

from .pyproject_configs import RobotpyBuildConfig
from .pkgcfg_provider import PkgCfgProvider
from .platforms import get_platform
from .wrapper import Wrapper


class Setup:
    """
        Hacky wrapper around setuptools because it's easier than copy/pasting
        this stuff to a million setup.py files
    """

    def __init__(self):
        self.root = abspath(os.getcwd())
        self.wrappers = []

        project_fname = join(self.root, "pyproject.toml")

        try:
            with open(project_fname) as fp:
                self.pyproject = toml.load(fp)
        except FileNotFoundError as e:
            raise ValueError("current directory is not a robotpy-build project") from e

        self.project_dict = self.pyproject.get("tool", {}).get("robotpy-build", {})
        try:
            self.project = RobotpyBuildConfig(**self.project_dict)
        except Exception as e:
            raise ValueError(
                f"robotpy-build configuration in pyproject.toml is incorrect"
            ) from e

        self.platform = get_platform()

    @property
    def base_package(self):
        return self.project.base_package

    @property
    def base_package_path(self):
        return join(self.root, *self.base_package.split("."))

    @property
    def git_dir(self):
        return join(self.root, ".git")

    def prepare(self):

        self.setup_kwargs = self.project_dict.get("metadata", {})
        self.setup_kwargs["zip_safe"] = False
        self.setup_kwargs["include_package_data"] = True
        self.setup_kwargs["python_requires"] = ">=3.6"

        self._generate_long_description()

        # get_version expects the directory to exist
        base_package_path = self.base_package_path
        os.makedirs(base_package_path, exist_ok=True)
        self.setup_kwargs["version"] = get_version(
            write_to=join(base_package_path, "version.py"), fallback_version="master"
        )

        self.pkgcfg = PkgCfgProvider()

        self._collect_wrappers()

        self.setup_kwargs["cmdclass"] = {
            "build_py": BuildPy,
            "build_dl": BuildDl,
            "build_gen": BuildGen,
            "build_ext": BuildExt,
        }
        if bdist_wheel:
            self.setup_kwargs["cmdclass"]["bdist_wheel"] = bdist_wheel
        for cls in self.setup_kwargs["cmdclass"].values():
            cls.wrappers = self.wrappers

        # We already know some of our packages, so collect those in addition
        # to using find_packages()
        packages = {w.package_name for w in self.wrappers}
        packages.update(find_packages())
        self.setup_kwargs["packages"] = list(packages)

    def _generate_long_description(self):
        readme_rst = join(self.root, "README.rst")
        readme_md = join(self.root, "README.md")
        if exists(readme_rst):
            with open(readme_rst) as fp:
                self.setup_kwargs["long_description"] = fp.read()

        elif exists(readme_md):
            self.setup_kwargs["long_description_content_type"] = "text/markdown"
            with open(readme_md) as fp:
                self.setup_kwargs["long_description"] = fp.read()

    def _collect_wrappers(self):

        ext_modules = []

        for package_name, cfg in self.project.wrappers.items():
            w = Wrapper(package_name, cfg, self)
            self.wrappers.append(w)
            self.pkgcfg.add_pkg(w)

            if w.extension:
                ext_modules.append(w.extension)

        if ext_modules:
            self.setup_kwargs["ext_modules"] = ext_modules

    def run(self):
        # assemble all the pieces and make it work
        _setup(**self.setup_kwargs)


def setup():
    s = Setup()
    s.prepare()
    s.run()



from distutils.core import Command
import os.path


class BuildDl(Command):

    command_name = "build_dl"
    description = "Downloads files"
    user_options = [
        ("build-base=", "b", "base directory for build library"),
        ("build-cache=", None, "build directory to cache downloaded objects"),
        ("src-unpack-to=", None, "build directory to unpack sources to"),
    ]
    wrappers = []

    def initialize_options(self):
        self.build_base = None
        self.build_cache = None
        self.src_unpack_to = None

    def finalize_options(self):
        self.set_undefined_options("build", ("build_base", "build_base"))
        if self.build_cache is None:
            self.build_cache = os.path.join(self.build_base, "cache")
        if self.src_unpack_to is None:
            self.src_unpack_to = os.path.join(self.build_base, "dlsrc")

    def run(self):
        for wrapper in self.wrappers:
            wrapper.on_build_dl(self.build_cache, self.src_unpack_to)
